/*****************************************************************************/
//	BitVector.h
/*****************************************************************************/

#ifndef _RdbBitVector_H_
#define _RdbBitVector_H_

#include <stdio.h>

/*****************************************************************************/
/*
	A bitvector is an array of bits whose length is given by the 'length'
	field. Bit 'i' is true iff:
	    words[i>>BIT_VECTOR_WORD_LOG] & (1<<(i & BIT_VECTOR_WORD-1)).
	If length<=BIT_VECTOR_WORD, words == &word0;
	otherwise words points to a malloc'd buffer.
*/
/*****************************************************************************/

#define BIT_VECTOR_WORD 32
#define BIT_VECTOR_WORD_LOG 5 /* log2(BIT_VECTOR_WORD) */

/*****************************************************************************/

typedef unsigned long BitWord;

/*****************************************************************************/

struct BitVector {
  private:
    inline const BitWord * words() const {
	return length <= BIT_VECTOR_WORD ? &u.word0 : u.words; 
    }
    inline BitWord * nc_words()	// returns a non-const version of words()
      { return (BitWord*)words(); }

  public:
    int length;
    union {
	BitWord	*words;	  // if length > BIT_VECTOR_WORD 
	BitWord	word0;
    } u;

    void init() { length = 0; u.word0 = 0; }
    void init(int len) { init(); setlen(len); }
    void init(const BitVector&);
    
  public:
    
    BitVector() { init(); }
    BitVector(int len) { init(len); }
    BitVector(int l, const BitWord w) { length = l; u.word0 = w; }
    BitVector(int l, BitWord* wp) { length = l; u.words = wp; }
    BitVector(const BitVector& bv) { init(bv); } // *real* Copy constructor.
    ~BitVector();

    void clear(); // Make all bits false, without adjusting length
    void setlen(int l); // Change length to 'l'
    void set(int index, int value = 1); // Set bit 'index' to 'value'

    // Return 1 iff bit 'index' is true.
    inline int test(int index) const {
	if (index >= length)
	  return 0;
	else if (length <= BIT_VECTOR_WORD) {
	    return (u.word0 & (1<<(index & BIT_VECTOR_WORD-1))) ? 1 : 0;
	}
	else {
	    return (u.words[index>>BIT_VECTOR_WORD_LOG] &
		    (1<<(index & BIT_VECTOR_WORD-1))) ? 1 : 0;
	}
    }

    int next(int i) const; // Return next true bit, after 'i'; or -1 if none.
    int count() const; // Return number of true bits.
    int any() const; // True if any bit is true.
    inline int len() const { return length; }
    void printon(FILE* file) const; // Print on file using format "110010"

    inline int operator[](int index) const { return test(index); }
    int operator==(const BitVector& b2) const;
				// True iff '*this' and 'b2' are equal.
    int operator&(const BitVector& b2) const;
				// True iff both args have common true bits.
    int implies(const BitVector& b2) const;
				// True iff all set bits in this are set in b2
    void operator=(const BitVector& b2);
    void or_bits(const BitVector *b);
				// Set all bits in this that are set in *b

};


#endif // _RdbBitVector_H_
