// *************************************************************************
//    HashSimpleRelation.C
// *************************************************************************

#include "HashSimpleRelation.h"

#include "Globals.h"
#include "CoralError.h"
#include "Unify.h"
#include "Debug.h"

#include "TupleList.h"

#ifdef ASYN_SOCK_REL
#include "Util.h"
#include "ClientProtocol.h"
#endif

// *************************************************************************
Cor_HashSimpleRelation::Cor_HashSimpleRelation( int arity,
            int delta, int buckets)
    : Cor_StorageRelation(arity, delta), index_set(this)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // NOTE: remember to update the code in empty_relation() and in 
    //       move_all_entries() when updating this code.

    Cor_BitVector *bv = new Cor_BitVector(arity);

    for (int i = 0; i < arity; i++) {
        // all arguments bound
        bv->set(i);
    }

      // LEAKS (in rare cases?) - SPF
    Cor_HashIndex *h1 = new Cor_HashIndex(this, bv, buckets); 

      // No one should have taken a pointer to this.
    delete bv;

    index_set.insert_index(h1);
}

// *************************************************************************
void Cor_HashSimpleRelation::empty_relation(int deleteTuples)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Empty out the TupleList
    tlist.free_all();

    // Empty out the indices
    for (int i = 1; i < index_set.num_indices(); i++) {
        index_set[i]->delete_all_tuples();
        index_set[i]->reset_index();
    }

    // Empty all tuples out of relation.  
    index_set[0]->delete_all_tuples(deleteTuples);
    index_set[0]->reset_index();

    // Recursively empty out the delta relations and add them to the 
    // free_reln_chain

    Cor_RelationLink *old;
    if (delta_indexed && delta_relations) {
        while(delta_relations->reln_chain) {
            Cor_Relation *reln = delta_relations->reln_chain->reln;
            old = delta_relations->reln_chain;
            delta_relations->reln_chain = old->next;

            reln->empty_relation(deleteTuples);    
            old->next = free_reln_list;
            free_reln_list = old;
        }
        delta_relations->reln_lastptr = NULL;
        delete delta_relations;
        delta_relations = NULL;
    }

    headMark.next = &headMark;
    headMark.prev = &headMark;

    count = 0;   
}

// *************************************************************************
int Cor_HashSimpleRelation::move_all_entries(Cor_StorageRelation *rel)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Moves all entries in the main relation of (this) to the relation rel.
    // This DOES NOT move entries that are in the delta_relations list or in 
    // the LinkedRelation list.  
    // In general it would be good to have this function
    // do that too, but right now this function is configured to only move
    // tuples from delta relations into the main relation.

    headMark.next = &headMark;
    headMark.prev = &headMark;

    // NOTE:  If relation count is present, reset it here.
    rel->count += count;

    index_set.move_all_entries(rel);
    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->reset_index();
    }
    count = 0;

    return(1);
}

// *************************************************************************
void Cor_HashSimpleRelation::free_indices()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    tlist.free_all();

    // Don't delete index 0, need that one yet.
    for (int i = 1; i < index_set.num_indices(); i++)
        delete ((Cor_HashIndex *) index_set[i]); 
}

// *************************************************************************
void Cor_HashSimpleRelation::free_delta_relations()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_RelationLink *curr, *prev = NULL;
    if (delta_indexed && delta_relations) {
        for (curr=delta_relations->reln_chain; curr; curr = curr->next) {
            delete curr->reln;
            delete prev;
            prev = curr;
        }
        delete prev;
        prev = NULL;
    }

    for (curr = free_reln_list; curr; curr = curr->next) {
        delete curr->reln;
        delete prev;
        prev = curr;
    }
    delete prev;

    free_reln_list = NULL;
}

// *************************************************************************
void Cor_HashSimpleRelation::free_all()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    headMark.next = &headMark;
    headMark.prev = &headMark;
    
    free_indices();

    free_delta_relations();

    Cor_HashIndex *index0 = (Cor_HashIndex *) index_set[0];
    ASSERT (index0 != NULL);

    index0->delete_all_tuples(!_borrow);
    delete (Cor_HashIndex *) index0;
}

// *************************************************************************
Cor_HashSimpleRelation::~Cor_HashSimpleRelation()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    free_all();
}

// *************************************************************************
void Cor_HashSimpleRelation::freeMark(Cor_RMark *mark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (mark == NULL)
        return;

    Cor_RMark *this_mark = mark;

    // Unlink mark
    this_mark->next->prev = this_mark->prev;
    this_mark->prev->next = this_mark->next;

    // add it to the free marks list
    this_mark->next = free_marks;
    free_marks = this_mark;

    // check if this was the first mark in the list of marks
    if (this_mark->prev != &headMark)
        return;

    // thus, if an "intermediate" mark is deleted, nothing is done!
    // this is fine for (!delta_indexed) relations, but with delta_indexed
    // relations, an alternative is to merge the delta relations 
    // corresponding to this mark with the "previous" relation.

    if (!delta_indexed) {

        int limit;

        // Move all linked tuples upto the new first mark to hash table.
        if (there_are_marks()) {
            limit = headMark.next->tindex;
            for (Cor_RMark *hmark = headMark.next; ; hmark = hmark->next) {
                if (hmark == &headMark)
                    break;
                if (hmark->tindex != limit)
                    break;
                hmark->tlist = &tlist;
                hmark->tindex = 0;
            }
        } else {
            limit = tlist.count();
        }

        for (int j=0; j < limit; j++) {
            Cor_Tuple *tuple = tlist.tuples[j];

            if (!insert_raw(tuple))
                delete tuple;
            else
                count++;

            tlist.tuples[j] = NULL;
        }

    } else { 
        Cor_RelationLink *limit;

        ASSERT((delta_relations!= NULL));
        ASSERT((delta_relations->reln_lastptr != NULL));
        
        if (there_are_marks()) 
            limit = headMark.next->rel_p;
        else
            limit = NULL;

        Cor_RelationLink *old;
        while (delta_relations->reln_chain != limit) {
            Cor_StorageRelation *reln = delta_relations->reln_chain->reln;
            old = delta_relations->reln_chain;
            delta_relations->reln_chain = old->next;

#            define DELTA_INDEX_HACK 
#            ifdef DELTA_INDEX_HACK
                // WARNING - we assume that the delta relation is a 
                // HashSimpleRelation.

                ((Cor_HashSimpleRelation*)reln)->move_all_entries(this);
#            else
                reln->add_all_tuples(this,0,0);
                reln->empty_relation();
#            endif

            old->next = free_reln_list;
            free_reln_list = old;
        }

        if (delta_relations->reln_chain == NULL)
            delta_relations->reln_lastptr = NULL;
    }
}

// *************************************************************************
void Cor_HashSimpleRelation::free_allMarks()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_RMark *hmark;

    while (headMark.next != &headMark) {
        hmark = headMark.next;
        freeMark(hmark);
    }
    ASSERT((headMark.prev == headMark.next));
}

// *************************************************************************
void Cor_HashSimpleRelation::initialize_iterator( Cor_TupleIterator& iterator )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_TRACE
        if (Cor_exEnv->dbg_get_next) {
            *(Cor_exEnv->trace_strm) << 
                "\nHashSimpleRelation::initialize_iterator: ";
            iterator.relation->name.print(NULL, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " ( ";
            iterator.arg_list.print(iterator.bindenv, Cor_exEnv->trace_strm);
            *(Cor_exEnv->trace_strm) << " )\n";
        }
#    endif
    iterator.stack_mark.get_mark();

    Cor_Term t1;

    (void) scan_count++;

    if (!delta_indexed) {
        if (iterator.start_mark) {
            // don't need to go to the hash table
            iterator.ipos = Scanning_New;
            iterator.bkt  = (iterator.start_mark)->tlist;
            iterator.ppos = (iterator.start_mark)->tindex;
        } else {
            // need to scan from hash table
            // find the appropriate index to be used
            iterator.index = index_set.find_index(iterator);
        }
    } else {
        if ( !iterator.in_delta()) {
            // Starting on main relation
            if (iterator.start_mark) {
                // go to the delta relations
                iterator.ipos = Scanning_Init;
                iterator.delta_rel_pos = (iterator.start_mark)->rel_p;
            } else {
                // need to scan from hash table
                // find the appropriate index to be used
                iterator.index = index_set.find_index(iterator);
            }
        } else {
            // In a delta relation
            iterator.index = index_set.find_index(iterator);
        }
    }
}

// *************************************************************************
//  Tries to find tuple in relation that unifies with 
//      <iterator.arg_list, iterator.bindenv>
//
//  On success:  
//     Resets iterator.no_match, sets iterator.tuple_env to
//     the environment created for the tuple, and returns the fetched 
//     tuple.  The tuple is NOT a copy, and should not be deleted or 
//     destructively modified where it is called from.  The same holds
//     for tuple.bindenv.  However, iterator.tuple_env (if it is non-null)
//     is a local copy and may be modified by the calling routine.
//     It may need to be disposed off by the calling routine once it has
//     served its purpose.
//
//     In general, DO NOT assume that a return value of NULL implies 
//     that no tuple was found - evaluable relations do not create tuples,
//     but merely modify iterator.bindenv.  Hence they always return null.
//
//  On failure:  
//      Sets iterator.no_match.
// *************************************************************************

#define FAIL_RETURN \
do { \
    iterator.ipos = Scanning_Done;     \
    iterator.set_no_match();    \
    return NULL;            \
} while(0)

Cor_Tuple *Cor_HashSimpleRelation::get_next_tuple(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#    ifdef DO_PROFILE
        if (Cor_T_Stack->count > 0) {
            (Cor_T_Stack->top()->moduledata->
                module_info->Cor_GlobalProfileInfo.getnexts)++;
        } else {
            (Cor_GlobalProfileInfo.getnexts)++;
        }
#    endif

    Cor_Tuple *result = NULL;
    iterator.tuple_env = NULL;

    if (Cor_exEnv->C_interrupt_raised)
        FAIL_RETURN;

    if ((iterator.ipos == Scanning_Init)) 
        initialize_iterator( iterator );
    else
        iterator.stack_mark.pop_to();

    if (!delta_indexed) {
        // do the old stuff with chain and everything
        if (iterator.ipos != Scanning_New) {

            // need to scan from hash table
            result = iterator.index->get_next_tuple(iterator);

            if (result != NULL) {
                // found something in hash table
                return(result); 
            } else {
                iterator.ipos = Scanning_New;
                iterator.bkt  = &tlist;
                iterator.ppos = 0;

                // this is required because we don't want 
                // HashIndex::get_next_tuple() to set iterator.ipos to 
                // Scanning_New at the end of search
            }
        }
    
        Cor_StackMark stack_mark;

        // need to scan from the chain now

        Cor_Tuple *tuple;

        while (iterator.ppos < iterator.bkt->count()) {

            if (! (tuple = iterator.bkt->tuples[iterator.ppos])) {
                (iterator.ppos)++;
                continue;
            }

            if (iterator.end_mark 
                    && (iterator.end_mark->tindex <= iterator.ppos) ) {
                tuple = NULL;
                break;
            }

            (iterator.ppos)++;

            if (tuple->is_deleted())
                continue;

            if (Cor_unify_literal(iterator, tuple, Cor_exEnv->dbg_get_next)) 
                return(tuple);

            stack_mark.pop_to();
        }

        iterator.ipos = Scanning_Done;
        iterator.set_no_match();
        stack_mark.pop_to();
        return(NULL);
    } else {
        // relation is delta-indexed

        //  WARNING: the following hack has been done.
        //  iterator not copied when calling delta relations.  so it
        //  has the same "relation" field value, but each time the index field
        //  changes based on the main/delta relations.
        //  note that iterator.get_next_tuple is NOT called, which
        //  would be wrong because the "relation" field value is not correct!!
        //
        //  The main/delta relations->get_next_tuple is called directly.
        //  The following ASSUMPTION is made:
        //  -- relation->get_next_tuple(iterator) does NOT use the 
        //     relation field of iterator.
        //  -- HashSimpleRelation::get_next_tuple satisfies this assumption.

        if (!iterator.in_delta()) {
            // Scanning from main hash table
            result = iterator.index->get_next_tuple(iterator);
            if (result != NULL)
                return(result);
            if (!delta_relations)
                FAIL_RETURN;

            // now go into the delta relations
            iterator.ipos = Scanning_Init;
            iterator.delta_rel_pos = delta_relations->reln_chain;
        }

        // delta_rel_pos has to have been set by this point.
        if (delta_relations != NULL) {
            // iterator scanning deltas, this == main relation
            for(;;) {
                // Loop till successful delta is found
                if (iterator.ipos == Scanning_Init) {
                    if (iterator.delta_rel_pos == NULL) {
                        FAIL_RETURN;
                    }
                    if (iterator.end_mark && (iterator.delta_rel_pos ==
                            (iterator.end_mark)->rel_p)) {
                        FAIL_RETURN;
                    }
                }
                result = iterator.delta_rel_pos->reln->get_next_tuple(iterator);
                if (result != NULL)
                    return(result);

                // Done with current delta relation.  Got to go to next delta.
                iterator.delta_rel_pos = iterator.delta_rel_pos->next;
                iterator.ipos = Scanning_Init;      
            } 
        } else {
            // iterator scanning deltas, this == delta relation
            return(iterator.index->get_next_tuple(iterator));
        }
    }
}

#undef FAIL_RETURN

// *************************************************************************
//  As above, except that it does not return the tuple, but instead
//  returns iterator.bindenv on success, and NULL on failure.
//  returns a fetched tuple
// *************************************************************************
Cor_BindEnv *Cor_HashSimpleRelation::get_next(Cor_TupleIterator& iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_Tuple *result = get_next_tuple(iterator);
    if (!result)
        return(NULL);
    return(iterator.bindenv);
}

// *************************************************************************
void Cor_HashSimpleRelation::release(Cor_TupleIterator& )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // Do nothing.  No state release is required
}

// *************************************************************************
int Cor_HashSimpleRelation::insert_raw( Cor_Tuple *tuple )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // inserts tuple into each index constructed for the relation
    if (!tuple || tuple->is_deleted()) {
        // tuple has been logically deleted
        return(0);
    }

    tuple->in_tlist = 0;

    for (int i = 0; i < index_set.num_indices(); i++) {
        index_set[i]->insert_tuple(tuple);
    }
    return(1);
}

// *************************************************************************
int Cor_HashSimpleRelation::is_subsumed(Cor_Tuple *newtuple,
        Cor_RMark *start_mark, Cor_RMark *end_mark)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int result;
    Cor_GenericIndex *index;

    if (! start_mark) {
        // need to go to the hash table
        // find the appropriate index to be used
        index = index_set.find_index(newtuple);
    } else
        index = NULL;

    if (!delta_indexed) {
        // do the old stuff with chain and everything
        if (index) {
            // need to scan from hash table
            result = index->is_subsumed(newtuple);
            if (result != 0) {
                // found something in hash table
                return(result); 
            }
        }

        Cor_Tuple *tuple;

        for (int j=0; j < tlist.count(); j++) {

              // Continue until we find a match
            if (end_mark && (end_mark->tindex < j))
                break;

            tuple = tlist.tuples[j];

            if (!tuple || tuple->is_deleted())
                continue;

            if (Cor_subsumes_tuple(tuple, newtuple, (DISPLAY_INSERTIONS)))
                return(1);
        }

        return( is_agg_subsumed(newtuple, start_mark, end_mark) );
    } else {
        // relation is delta-indexed

        if (index) {
            // Need to look at index for this relation.
            result = index->is_subsumed(newtuple);
            if (result != 0) {
                return(result);
            }
        }

        if (delta_relations != NULL) {
            Cor_RelationLink *delta_rel_pos = delta_relations->reln_chain;

            if (start_mark) { // Find first delta relation to be used
            while((start_mark)->rel_p != delta_rel_pos &&
                delta_rel_pos != NULL)
                delta_rel_pos = delta_rel_pos->next;
            }

            for(;;) {     // Loop till successful delta is found
                if (!delta_rel_pos) {
                    // End of delta_relations chain
                    return(is_agg_subsumed(newtuple, start_mark, end_mark));
                }

                if (end_mark && (delta_rel_pos == (end_mark)->rel_p)) {
                    return(is_agg_subsumed(newtuple, start_mark, end_mark));
                }

                result = delta_rel_pos->reln->is_subsumed(newtuple);
                if (result != 0) {
                    return(result);
                }

                // Done with current delta relation.  Got to go to next delta.
                delta_rel_pos = delta_rel_pos->next;
            } 
        }
        return(is_agg_subsumed(newtuple, start_mark, end_mark));
    }
}

// *************************************************************************
int Cor_HashSimpleRelation::insert_tuple(Cor_Tuple *tuple)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (there_are_marks()) {
        if (delta_indexed) {
            tuple->in_tlist = 0;
            return(delta_relations->insert_tuple(tuple));
        } else {
            // PRAVEEN :: this here is a problem !!!
            count++;
            //return(tlist_insert_tuple(tuple));
            if (!tuple->is_deleted()) {
                tlist.push(UnknownHashValue,tuple);
                tuple->in_tlist = 1;
                return(1);
            } else
                return(0);
        }
    } else {
        count++;
        tuple->in_tlist = 0;
        return(insert_raw(tuple));
    }

    // This section is now unreachable.

    //    _hash = UnknownHashValue; 
    // to ensure set is re-hashed, since it changed.
    // this fix doesn't work, consistency of hash-cons table may be violated!

    // return StorageRelation::insert_tuple(tuple);
    return(1);
}

// *************************************************************************
void Cor_HashSimpleRelation::print_name(ostream *out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *out << name.symbol() << "/" << arity() << "  : (base)    ";
    index_set.print_adornments(out);
    *out << ": count " << count << endl;
}

// *************************************************************************
void Cor_HashSimpleRelation::show_internals(ostream& out)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef    DEBUG
    out << name.symbol() << "/" << arity() << "  : HashSimple    ";
    index_set.print_adornments(&out);
    out << ": count " << count << endl;

    // Now show the internals of each index and delta
    // spf - not done yet.

    if (!delta_indexed) {
        out << "not Delta Indexed: \n";
        out << tlist << endl;
    } else {
        out << "Delta Indexed: \n";
        if (delta_relations) {

            for (Cor_RelationLink *rlink=delta_relations->reln_chain;
                    rlink; rlink = rlink->next) {

                out << "Showing a delta of " << this << endl;
                rlink->reln->show_internals(out);
            }
        }
    }

    for (int j=0; j < index_set.num_indices(); j++) {
        out << "Index " << j << endl;
        index_set[j]->show_all_tuples(out);
    }

#endif 
}

// *************************************************************************
#ifdef ASYN_SOCK_REL
void Cor_HashSimpleRelation::get_tuples(void)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    char Buf[1000];

    int endOfResult = 0;
   
    if (!cs) { 
        _delta_done = 1;
        return;
    }

    while (((Cor_sockbuf*)cs->rdbuf())->more()) {
         if (!(cs->gets(Buf, 1000)))
                break;

         endOfResult = 0;
         
         if (Buf[4] == ':'){
             if (strncmp(Buf,CP_EOQ, 5) == 0) {
                 if (--open_queries == 0)
                     endOfResult = 1;
             }
             else if (strncmp(Buf,CP_CERR,5) == 0) {
                 *(Cor_exEnv->out_strm) << (&(Buf[5]));
             }
             else if (strncmp(Buf,CP_MESG,5) == 0) {
                 *(Cor_exEnv->out_strm) << (&(Buf[5]));
             }
             else if (strncmp(Buf,CP_QID, 5) == 0) {
                // Only for multiplexing multiple relations onto a
                // single socket?
             }
             else {
                 *(Cor_exEnv->err_strm)
                        << ("Warning, unimplemented protocol header.\n");
                 *(Cor_exEnv->out_strm) << Buf;
             }
         } else {

             // Otherwise, it must be data.

             Cor_CArg relname;
             int ar;
             Cor_Tuple *tuple = Cor_parseFact(Buf,relname, &ar);

             // SPF
             // cout << "parsed fact: " << relname << *tuple << endl;

             if (tuple)
                 Cor_Relation::insert_new(tuple);
         }
    }

    if (endOfResult)
        _delta_done = 1;
}
#endif // ASYN_SOCK_REL
// *************************************************************************
