// *************************************************************************
//    SccInfo.C
// *************************************************************************

#include "SccInfo.h"

#include "Globals.h"
#include "Literal.h"
#include "BuiltinSymbols.h"
#include "Compile.h"
#include "CompileRule.h"
#include "Execute.h"
#include "Unparse.h"
#include "Debug.h"

#ifdef ASYN_SOCK_REL
#include "AsynSockRelation.h"
#endif

extern void Cor_EndOfScc(Cor_SccInfo *scc, int iters, ostream *outf);

// *************************************************************************
Cor_SccInfo::Cor_SccInfo(Cor_QFMaterializedModuleData *md, Cor_Scc *scc,
    Cor_Predicate *ext_magicpred, int id)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
#ifdef ASYN_SOCK_REL
    SockRelPreds = new Cor_PredicateList();
#endif
    
    module = md;
    this->scc = scc;
    scc->sccInfo = this;
    scc_id = id;
    this->createSNStructs(ext_magicpred);

    rec_rule_start = 0;
    pred_success_arr = 0;
}

// *************************************************************************
Cor_SccInfo::~Cor_SccInfo()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    if (numPreds) {
        delete [] pred_ptrs;
        delete [] offset_arr;
        delete [] pred_success_arr;
        delete [] sn_rules;
        delete [] sn_offset_arr;
    }

#ifdef ASYN_SOCK_REL
    delete SockRelPreds;
#endif
}

// *************************************************************************
int Cor_SccInfo::allocateRuleStructures(Cor_Predicate * /* ext_magicpred */)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // First, allocate enough memory to hold all seminaive rule structures
    numSNRules = 0;
    Cor_Predicate *pred; Cor_Clause *clause;
    int i;

    for (pred = scc->preds; pred != NULL; pred = pred->next) {
        for (clause = pred->clauses; clause != *pred->clause_tail;
            clause = clause->pnext) {

            if (!clause->seen)
                continue;

            if (!clause->recursive) {
                numSNRules++;
                continue;
            }

            for (int ipred = 0; ipred < clause->num_preds; ipred++) {
                // Treat negated literals as base
                if (clause->preds[ipred]->negated)
                    continue;

                // The following is to special case the creation of the
                // seminaive structures for the case of Ordered Search.
                // Three special things need to be done
                // (1) `not p' is always treated as base, even if p is derived
                // (2) done_m_p is treated as derived even though there are no
                //        rules for it.
                // (3) any literal following a done_m_p is treated as base,
                //        even if p is derived. (this also covers case 1)

                if (module->module_info->UseOrdSearch) {
                    if ((ipred > 0) &&
                        clause->preds[ipred-1]->predicate->Cor_is_mp_done())
                            continue;

                    // Treat mp_done literals as derived
                    if (clause->preds[ipred]->predicate->Cor_is_mp_done())
                        clause->preds[ipred]->predicate->scc = scc;
                }

#ifdef ASYN_SOCK_REL
                if (!clause->preds[ipred]->predicate->asyn_sock()) {
#endif 
                    if (clause->preds[ipred]->predicate->scc != scc) {
                        continue;
                    }
#ifdef ASYN_SOCK_REL
                } 
#endif
                numSNRules++;
            }
        }
    }

    sn_rules = new Cor_SemiNaive[numSNRules]; 
    for (i = 0; i < numSNRules; i++)
        sn_rules[i].rule_id = i;

    sn_offset_arr = new int[numSNRules];

    // find the number of predicates in this Scc
    for (numPreds = 0, pred = scc->preds; pred != NULL;
        ++numPreds, pred = pred->next);

    // now fill in the information about lower Sccs

    if (module->numSccs == 1) {             // this is the first Scc
        num_lowerPreds = 0;
    } else {
          // add the numPreds and num_lowerPreds of the previous scc
        ASSERT((module->numSccs > 1));
        num_lowerPreds =
            module->sccArray[module->numSccs-2]->num_lowerPreds +
            module->sccArray[module->numSccs-2]->numPreds;
    }

#ifdef ASYN_SOCK_REL
    numSockRelPreds = 0;
    for (pred = scc->preds; pred != NULL; pred = pred->next) {
        for (clause = pred->clauses; clause != *pred->clause_tail;
            clause = clause->pnext) {

            for (int ipred = 0; ipred < clause->num_preds; ipred++) {
                if (clause->preds[ipred]->predicate->asyn_sock() ) {  
                    SockRelPreds->reset();
                    Cor_Predicate *sock_pred = SockRelPreds->get_next();
                    while (sock_pred) {
                        if (sock_pred == clause->preds[ipred]->predicate)
                            break;
                        sock_pred = SockRelPreds->get_next();
                    }
                    if (!sock_pred) {
                        SockRelPreds->append(clause->preds[ipred]->predicate);
                        numSockRelPreds++;
                    }
                }
            }
        }
    }

    pred_ptrs =
        new Cor_Predicate* [numPreds + num_lowerPreds + numSockRelPreds];
    offset_arr = new int [numPreds + num_lowerPreds + numSockRelPreds];
#else
    // assign the predicate pointers and the offset array
    pred_ptrs = new Cor_Predicate* [numPreds];
    offset_arr = new int[numPreds];
#endif
    
    for (i = 0, pred = scc->preds; pred != NULL; i++,pred = pred->next) {
        pred_ptrs[i] = pred;
        // allow at least one iteration
    }

#ifdef ASYN_SOCK_REL
    int j;
    // get the predicate pointers from the lower scc
    if (module->numSccs > 1) {
        Cor_Predicate **lower_pred_ptrs =
            module->sccArray[module->numSccs-2]->pred_ptrs;
        ASSERT((i == numPreds));
        for (j=0; j < num_lowerPreds; j++, i++) {
            pred_ptrs[i] = lower_pred_ptrs[j];
        }
    } else { // module->numSccs == 1
        ASSERT((num_lowerPreds <= 1));
        if (num_lowerPreds == 1) {
            pred_ptrs[numPreds] = NULL;
        }
    }
#endif
    
#ifdef ASYN_SOCK_REL
    SockRelPreds->reset();
    for (j = 0; j < numSockRelPreds; i++, j++) 
        pred_ptrs[i] = SockRelPreds->get_next(); 
#endif
    
    return(numSNRules);
}

// *************************************************************************
//  Routine that creates the seminaive structures in each Scc structure 
// *************************************************************************

void Cor_SccInfo::createSNStructs(Cor_Predicate *ext_magicpred)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    // allocate sn_rules, pred_ptrs, offset_arr
    allocateRuleStructures(ext_magicpred);

    // Now fill in values in seminaive rule structures
    Cor_Predicate *pred; Cor_Clause *clause;

    int cur_snr = -1, i ,j;

    // There is a correspondence between the predicate positions in 
    // this->pred_ptrs, and the offsets in the frame given by 
    // this->offset_arr of those predicates in rm_Arr.
    
    for (i=0; i<numPreds; i++)
        offset_arr[i] = Cor_get_offset(module->F, pred_ptrs[i]);

#ifdef ASYN_SOCK_REL
    for (i=numPreds; i < numPreds+num_lowerPreds; i++) 
        offset_arr[i] = Cor_get_offset(module->F, pred_ptrs[i]);
    
    for (; i < numPreds + num_lowerPreds + numSockRelPreds; i++)
        offset_arr[i] = Cor_get_offset(module->F, pred_ptrs[i]);
#endif
    
    // Consider non-recursive clauses in this scc first.
    for (j=0, pred = scc->preds; pred != NULL; ++j, pred = pred->next)
    {
        for (clause = pred->clauses; clause != *pred->clause_tail;
            clause = clause->pnext)
        {
            if (!clause->seen)
                continue;
            if (clause->recursive)
                continue;

            // we know it is a non-recursive clause by now

            cur_snr++;
            ASSERT((cur_snr < numSNRules));
            sn_offset_arr[cur_snr] = j;
            sn_rules[cur_snr].scc = this;

            // the following function encapsulates the initialization of
            // a non-recursive seminaive rule. This is done so that 
            // imperative rules from the coral prompt can also be
            // initialized using the same code.
            Cor_compile_nonrec_rule(sn_rules[cur_snr], clause,
                    module->ruleArray->rInfoPtrs[clause->rule_number]);
        }
    }
    
    // Now handle recursive rules

    if (scc->recursive) {     // If there are recursive rules:

        // now generate recursive SemiNaive rules from recursive rules
        for (j=0, pred = scc->preds; pred != NULL; j++, pred = pred->next)
        {
            for (clause = pred->clauses; clause != *pred->clause_tail;
                clause = clause->pnext)
            {
                if (!clause->seen)
                    continue;
                if (!clause->recursive)
                    continue;

                int ipred, jpred, i;
                for (ipred = 0; ipred < clause->num_preds; ipred++) {
                    //
                    //    Treat negated literals as base !!
                    //
                    if (clause->preds[ipred]->negated) continue;

                    // with Ordered Search, have to special case the 
                    // negated literals as well as the grouped literal.
                    // the check below assumes that both follow a 
                    // done_m_p literal.

                    if (module->module_info->UseOrdSearch) {
                        if ((ipred > 0)
                                 && (clause->preds[ipred-1]->
                                    predicate->Cor_is_mp_done()))
                            continue;
                    }

                    // don't need to special case for mp_done since it is
                    // covered by the next if condition

#ifdef ASYN_SOCK_REL
                    if (!clause->preds[ipred]->predicate->asyn_sock()) {
#endif
                        if (clause->preds[ipred]->predicate->scc != scc)
                            continue;
#ifdef ASYN_SOCK_REL
                    }
#endif
                    
                    cur_snr++; 
                    ASSERT((cur_snr < numSNRules));
                    i = Cor_has_fail_in_body(clause);
                    if (i >= 0) {
                        sn_rules[cur_snr].first = i;
                    } else {
                        sn_rules[cur_snr].first = ipred;
                    }            
                    sn_rules[cur_snr].recursive = 1;
                    sn_rules[cur_snr].scc = this;
                    sn_offset_arr[cur_snr] = j;
                    sn_rules[cur_snr].Add_Rule_Info(
                            module->ruleArray->rInfoPtrs[clause->rule_number],
                            clause);
                    
                    for (jpred = 0; jpred < clause->num_preds; jpred++) {
                        Cor_Literal *bodypred = clause->preds[jpred];
                        
                        if (ipred == jpred) {
                            sn_rules[cur_snr].markOffsets[2*jpred] = 
                                    Cor_Find_RMark_Offset(this,
                                    bodypred->predicate);
                        } else {
                            sn_rules[cur_snr].markOffsets[2*jpred] = -1;
                        }
                        
#ifndef ASYN_SOCK_REL
                        if (bodypred->predicate->scc != scc)
                            sn_rules[cur_snr].markOffsets[2*jpred+1] = -1;
#else
                        if (!bodypred->predicate->asyn_sock() &&
                                bodypred->predicate->scc != scc)
                            sn_rules[cur_snr].markOffsets[2*jpred+1] = -1;
#endif
                        else {
                            int tmp = Cor_Find_RMark_Offset(this,
                                bodypred->predicate);
                            if (ipred > jpred)
                                sn_rules[cur_snr].markOffsets[2*jpred+1] =tmp;
                            else
                                sn_rules[cur_snr].markOffsets[2*jpred+1] =tmp+1;
                        }
                    }
                }
            }
        }
    }
    // when sn_rule generation is done, add labels to them
    for (i = 0; i<=cur_snr; i++)
            sn_rules[i].Add_label();

}

// *************************************************************************
void Cor_SccInfo::display_annotation_info(Cor_Table *F) 
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    CoralStream *file = Cor_exEnv->trace_strm;
    Cor_PredAnnotations *cur;
    Cor_QFModuleData *md = F->moduledata;
    
    for(int i=0; i < numPreds; i++) {
        Cor_CArg predname = F->rels[offset_arr[i]].relPtr->name;
        
        for(cur= md->agg_sel_annotations; cur; cur=cur->next()) {
            if (cur->pred.equals(predname)) {
                *file << "@" << Cor_AggSelectionSymbol->symbol()
                      << "    ";
                cur->print(file);
                *file << ".\n";
            }
        }

        *(Cor_exEnv->trace_strm)
            << F->rels[offset_arr[i]].relPtr->name.symbol()
            << ": subsumption checking - "
            << (F->rels[offset_arr[i]].relPtr->check_subsum? "on" : "off")
            << " \n\n";
        /* PRAVEEN:: Add the index information dumping routines here. */
    }
    *file << "---------------------------------------\n\n";
}

// *************************************************************************
Cor_RMarkArray *Cor_SccInfo::setupSolve(Cor_Table *F)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int sn_count;
    
    // Create array of RMarks
    // By default, all RMark *s are NULL 
    Cor_RMarkArray *rm_Arr = NULL;

    rm_Arr = F->module_env->scc_envs[this->scc_id]->rm_Arr;
    
    if ((DISPLAY_RULES(module)) && (!sn_rules[0].recursive) &&
        (!Cor_exEnv->C_quiet_mode_default))
            Cor_PrintSccNonRecursiveRules(this, Cor_exEnv->trace_strm);
    
    // For Basic SemiNaive, evaluate non-recursive rules

    for (sn_count = 0; sn_count < numSNRules; sn_count++) {
        if (sn_rules[sn_count].recursive)
            break;
        F->moduledata->curr_rInfo = sn_rules[sn_count].rInfo;
        sn_rules[sn_count].evaluate(F, rm_Arr);
    }

    rec_rule_start = sn_count;

    // Check for per_iteration aggregate selectors on non-recursive rules.
    for (int i=0; i < numPreds; i++) {
        Cor_StorageRelation *rel =
                (Cor_StorageRelation*) F->rels[offset_arr[i]].relPtr;

          // Does this relation have an aggregate selection on it?
        if (rel->agg_sel_info)
            rel->per_iteration_aggregation(NULL,NULL);
    }

    if (sn_count == numSNRules)
        return rm_Arr;


    // Note on RMarks for recursive rules. For each predicate, an 'old' RMark
    // and a 'new' RMark are created at consecutive positions in the RMarkArray.
    // The order in which these RMarks are created defines their positions in
    // the array, and this fact is used by Find_RMark_Offset() .

    if (DISPLAY_RULES(module) && !Cor_exEnv->C_quiet_mode_default)
        Cor_PrintSccRecursiveRules(this, rec_rule_start, Cor_exEnv->trace_strm);

    return rm_Arr;
}

// *************************************************************************
//  Performs one iteration over the recursive rules and returns 1 if the
//  iteration should continue, and 0 if the iterative loop should be exited.
// *************************************************************************

int Cor_SccInfo::iterateRecursiveRules(Cor_Table *F, Cor_RMarkArray *rm_Arr,
    int /* loop_iter_num */ )
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int new_tuples = 0;
    int sn_count;

      // Basic SemiNaive
    int i, j;
    for (i=0, j =1; i < numPreds; i++, j += 2) {
        rm_Arr->markList[j] =
            F->rels[offset_arr[i]].relPtr->getMark();
    }
        
    for (sn_count = rec_rule_start; sn_count < numSNRules; sn_count++) {
        Cor_SemiNaive *cur_sn_rule = &(sn_rules[sn_count]);
        F->moduledata->curr_rInfo = cur_sn_rule->rInfo;
        new_tuples += cur_sn_rule->evaluate(F, rm_Arr);
    }

    new_tuples += slide_deltas(F,rm_Arr,0);
   
    Cor_GoalNode *g_node; 

    if (new_tuples) {
        return(1);
    } else {
        // There are a bunch of special cases in which the number of
        // new_tuples is == 0, but there is still some computation to go.
        // However, if the value of new_tuples is non_zero, then we can
        // return right away.


        if (!module->module_info->UseOrdSearch)
            return(0);

        // Ordered Search
        // have reached a fixpoint; need to get unmarked tuple 
        // from context, and put it in the appropriate relation

        int any_new_fact_in_done = 0;
        g_node = F->context->get_unmarked_goal(any_new_fact_in_done);
        
        // exit condition from Ordered Search
        if (g_node == NULL) {
            if (!any_new_fact_in_done)
              return(0);
            else
              return(2);
        }

        g_node->owner_rel->insert_tuple(g_node);
        // NOTE: should remove g_node from 
        // g_node->owner_rel->for_ordsearch's index
        return(2);
    }
}

// *************************************************************************
//  Moves the deltas forward, freeing the old ones.
//  Return the net number of tuples added/removed from the delta.
// *************************************************************************
int Cor_SccInfo::slide_deltas(Cor_Table *F, Cor_RMarkArray *rm_Arr,
    int free_all)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int i,j;
    int ret = 0;

      // Need to check in case we have aggregation that needs to occur
      // at each iteration.
    for (i=0, j=0; i < numPreds; i++, j+=2) {
        Cor_StorageRelation *rel =
                (Cor_StorageRelation*) F->rels[offset_arr[i]].relPtr;

          // Does this relation have an aggregate selection on it?
        if (rel->agg_sel_info) {
            ret += rel->per_iteration_aggregation(
                       rm_Arr->markList[j],rm_Arr->markList[j+1]);
        }
    }


    for (i=0, j = 0; i < numPreds; i++, j+=2) {
        F->rels[offset_arr[i]].relPtr->freeMark(rm_Arr->markList[j]);
    }

    int numpreds2 = 2*numPreds;

    if (free_all)
        for (i=0; i < rm_Arr->array_size; i++) {
            rm_Arr->markList[i] = NULL;
        }
    else
        for (i=0, j=1; i < numpreds2; i+=2, j+=2) {
            rm_Arr->markList[i] = rm_Arr->markList[j];
            rm_Arr->markList[j] = NULL;
        }

    return(ret);
}

// *************************************************************************
//  Iterator will be non-null ONLY if this is a top-level scc and
//  staggered evaluation is desired on it.

void Cor_SccInfo::Solve(Cor_Table *F, Cor_TupleIterator *iterator)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    Cor_RMarkArray *rm_Arr;
    int i;
    
    // if staggered evaluation is not being performed, or this is the
    // not the top level scc, or this is the first time the top level 
    // scc is being executed

#ifdef    DEBUG
    //cout << "Entering Solve for SCC: " << this << endl;
#endif

    if (!iterator || !(iterator->state->rm_Arr)) {
        rm_Arr = setupSolve(F);

        if (iterator) {
            // first time this top-level scc is being executed
            iterator->state->rm_Arr = rm_Arr;
        }
    } else {
        rm_Arr = iterator->state->rm_Arr;
    }


    if (rec_rule_start == numSNRules) {
        if (rm_Arr) {
            for (i = 0; i < rm_Arr->array_size; i++) {
                rm_Arr->markList[i] = NULL;
            }
        }

        if (iterator)
            iterator->state->rm_Arr = NULL;

#       ifdef DO_PROFILE
            if ((Cor_exEnv->GlobalRelationOptions | module->rel_options) &
                REL_PROFILE_STATS)
                    Cor_EndOfScc(this, 0, Cor_exEnv->trace_strm);
#       endif
        return;
    }

      // loop_iter_num was for SaveModule, it isn't being used anymore.
    int loop_iter_num = 0;
    int ret_val = 1;

    // Evaluate recursive rules
    if (iterator) {
        // if lazy evaluation is being used, then the iter number is
        // actually stored in the iterator's state, and not in the variable
        // loop_iter_num.
        for(;ret_val;) {
            if ((ret_val = iterateRecursiveRules(F, rm_Arr, ++loop_iter_num))
                  && ret_val != 2)
                    return;
        }
    } else {
        while ((ret_val = iterateRecursiveRules(F, rm_Arr, ++loop_iter_num)))
            ;
    }
    
    // since the loop is going through at least once, these marks 
    // will not be the global old marks 

#ifndef ASYN_SOCK_REL
    slide_deltas(F,rm_Arr,1);
#endif
    
    if (iterator)
        iterator->state->rm_Arr = NULL;

#    ifdef DO_PROFILE
        if ((Cor_exEnv->GlobalRelationOptions | module->rel_options) &
            REL_PROFILE_STATS)
                Cor_EndOfScc(this, loop_iter_num, Cor_exEnv->trace_strm);
#    endif
}

// *************************************************************************
void Cor_SccInfo::print(ostream *fd, int print_preds, int print_text, 
            int print_snrules)
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    *fd << "Scc #" << scc_id << ": \n";
    scc->print(fd, print_preds, print_text);
    if (print_snrules) {
      for (int i=0; i<numSNRules; i++)
        sn_rules[i].print(fd);
    }
}
// *************************************************************************

#ifdef ASYN_SOCK_REL
int Cor_SccInfo::external_flag_on()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    for (int i = 0; i <= scc_id; i++)
        if (module->sccArray[i]->numPendingSockRel())
            return 1;
    return 0;
}

// *************************************************************************

int Cor_SccInfo::numPendingSockRel()
{
#ifdef    DO_CALL_STACK
    CallStk_Mark    callstk_mark(__FUNCTION__,__FILE__,__LINE__);
#endif
    int num = 0;
    SockRelPreds->reset();
    for (int i = 0; i < numSockRelPreds; i++) {
        if (!SockRelPreds->get_next()->delta_done())
            num++;
    }

    return num;
}    
#endif

// *************************************************************************
